@@ -5,7 +5,7 @@ import functools
 import inspect
 import io
 import tokenize
-from typing import Callable, Dict, List, Optional, Type
+from typing import Callable, Dict, Generic, Hashable, List, Optional, Type
 
 import docstring_parser
 from typing_extensions import get_origin, is_typeddict
@@ -117,7 +117,7 @@ def get_class_tokenization_with_field(
     for search_cls in classes_to_search:
         # Inherited generics seem challenging for now.
         # https://github.com/python/typing/issues/777
-        assert get_origin(search_cls) is None
+        assert search_cls is Generic or get_origin(search_cls) is None
 
         try:
             tokenization = _ClassTokenization.make(search_cls)  # type: ignore
@@ -190,20 +190,36 @@ def get_field_docstring(cls: Type, field_name: str) -> Optional[str]:
     #
     # In this case, 'Optimizer hyperparameters' will be treated as the docstring for all
     # 3 fields.
+
+    # Get first line of the class definition, excluding decorators. This logic is only
+    # needed for Python >= 3.9; in 3.8, we can simply use
+    # `tokenization.tokens[0].logical_line`.
+    classdef_logical_line = -1
+    for token in tokenization.tokens:
+        if token.content == "class":
+            classdef_logical_line = token.logical_line
+            break
+    assert classdef_logical_line != -1
+
     comments: List[str] = []
     current_actual_line = field_data.actual_line - 1
     while current_actual_line in tokenization.tokens_from_actual_line:
         actual_line_tokens = tokenization.tokens_from_actual_line[current_actual_line]
-        current_actual_line -= 1
 
         # We stop looking if we find an empty line.
         if len(actual_line_tokens) == 0:
             break
 
+        # We don't look in the first logical line. This includes all comments that come
+        # before the end parentheses in the class definition (eg comments in the
+        # subclass list).
+        if actual_line_tokens[0].logical_line <= classdef_logical_line:
+            break
+
         # Record single comments!
         if (
             len(actual_line_tokens) == 1
-            and actual_line_tokens[0].token_type == tokenize.COMMENT
+            and actual_line_tokens[0].token_type is tokenize.COMMENT
         ):
             (comment_token,) = actual_line_tokens
             assert comment_token.content.startswith("#")
@@ -212,12 +228,22 @@ def get_field_docstring(cls: Type, field_name: str) -> Optional[str]:
             # Comments should be contiguous.
             break
 
+        current_actual_line -= 1
+
     if len(comments) > 0:
         return "\n".join(reversed(comments))
 
     return None
 
 
+_builtins = set(
+    filter(
+        lambda x: isinstance(x, Hashable),  # type: ignore
+        __builtins__.values(),  # type: ignore
+    )
+)
+
+
 def get_callable_description(f: Callable) -> str:
     """Get description associated with a callable via docstring parsing.
 
@@ -226,6 +252,9 @@ def get_callable_description(f: Callable) -> str:
     these docstrings."""
     f, _unused = _resolver.resolve_generic_types(f)
 
+    if _resolver.unwrap_origin(f) in _builtins:
+        return ""
+
     # Note inspect.getdoc() causes some corner cases with TypedDicts.
     docstring = f.__doc__
     if (
