@@ -1,29 +1,13 @@
 """Interface for generating `argparse.ArgumentParser()` definitions from callables."""
-
 from __future__ import annotations
 
 import argparse
-import collections.abc
 import dataclasses
-import inspect
 import itertools
-from typing import (
-    Any,
-    Callable,
-    Dict,
-    Hashable,
-    List,
-    Optional,
-    Set,
-    Type,
-    TypeVar,
-    Union,
-    cast,
-)
+from typing import Any, Callable, Dict, List, Optional, Set, Type, TypeVar, Union, cast
 
 import termcolor
-import typing_extensions
-from typing_extensions import get_args, get_origin, is_typeddict
+from typing_extensions import get_args, get_origin
 
 from . import (
     _argparse_formatter,
@@ -38,59 +22,6 @@ from . import (
 T = TypeVar("T")
 
 
-_known_parsable_types = set(
-    filter(
-        lambda x: isinstance(x, Hashable),  # type: ignore
-        itertools.chain(
-            __builtins__.values(),  # type: ignore
-            vars(typing_extensions).values(),
-            vars(collections.abc).values(),
-        ),
-    )
-)
-
-
-def is_possibly_nested_type(typ: Any) -> bool:
-    """Heuristics for determining whether a type can be treated as a 'nested type',
-    where a single field has multiple corresponding arguments (eg for nested
-    dataclasses or classes).
-
-    Examples of when we return False: int, str, List[int], List[str], pathlib.Path, etc.
-    """
-
-    typ = _resolver.unwrap_origin(typ)
-
-    # Nested types should be callable.
-    if not callable(typ):
-        return False
-
-    # Known parsable types: builtins like int/str/float/bool, collections, annotations.
-    if typ in _known_parsable_types:
-        return False
-
-    # Simple heuristic: dataclasses should be treated as nested objects and are not
-    # parsable.
-    if dataclasses.is_dataclass(typ):
-        return True
-
-    # TypedDict types can be unpacked.
-    if is_typeddict(typ):
-        return True
-
-    # Nested types like nested (data)classes should have fully type-annotated inputs. If
-    # any inputs are unannotated (for example, in the case of pathlib.Path), we can
-    # assume the type is not nested.
-    try:
-        sig = inspect.signature(typ)
-    except ValueError:
-        return False
-    for param in sig.parameters.values():
-        if param.annotation is inspect.Parameter.empty:
-            return False
-
-    return True
-
-
 @dataclasses.dataclass(frozen=True)
 class ParserSpecification:
     """Each parser contains a list of arguments and optionally some subparsers."""
@@ -117,6 +48,7 @@ class ParserSpecification:
 
         # Resolve generic types.
         f, type_from_typevar = _resolver.resolve_generic_types(f)
+        f = _resolver.narrow_type(f, default_instance)
         if parent_type_from_typevar is not None:
             for typevar, typ in type_from_typevar.items():
                 if typ in parent_type_from_typevar:
@@ -140,40 +72,30 @@ class ParserSpecification:
         helptext_from_nested_class_field_name = {}
         subparsers_from_name = {}
 
-        for field in _fields.field_list_from_callable(
+        field_list = _fields.field_list_from_callable(
             f=f, default_instance=default_instance
-        ):
+        )
+        for field in field_list:
             field = dataclasses.replace(
                 field,
-                typ=type_from_typevar.get(field.typ, field.typ),  # type: ignore
+                typ=_resolver.type_from_typevar_constraints(
+                    type_from_typevar.get(  # type: ignore
+                        field.typ,
+                        field.typ,
+                    )
+                ),
             )
             if isinstance(field.typ, TypeVar):
-                if field.typ.__bound__ is not None:
-                    # Try to infer type from TypeVar bound.
-                    field = dataclasses.replace(field, typ=field.typ.__bound__)
-                elif len(field.typ.__constraints__) > 0:
-                    # Try to infer type from TypeVar constraints.
-                    field = dataclasses.replace(
-                        field, typ=Union.__getitem__(field.typ.__constraints__)  # type: ignore
-                    )
-                else:
-                    # Found an unbound TypeVar. This could be because inheriting from
-                    # generics is currently not implemented. It's unclear whether this is
-                    # feasible, because generics are lost in the mro:
-                    # https://github.com/python/typing/issues/777
-                    raise _instantiators.UnsupportedTypeAnnotationError(
-                        f"Field {field.name} has an unbound TypeVar: {field.typ}. Note"
-                        " that inheriting from generics is currently not implemented."
-                        " It's unclear whether this is feasible, because generics are"
-                        " lost in the mro: https://github.com/python/typing/issues/777"
-                    )
+                raise _instantiators.UnsupportedTypeAnnotationError(
+                    f"Field {field.name} has an unbound TypeVar: {field.typ}."
+                )
 
             # (1) Handle Unions over callables; these result in subparsers.
             subparsers_attempt = SubparsersSpecification.from_field(
                 field,
                 type_from_typevar=type_from_typevar,
                 parent_classes=parent_classes,
-                prefix=prefix + field.name + _strings.NESTED_FIELD_DELIMETER,
+                prefix=_strings.make_field_name([prefix, field.name]),
                 avoid_subparsers=avoid_subparsers,
             )
             if subparsers_attempt is not None:
@@ -189,24 +111,27 @@ class ParserSpecification:
                     continue
                 else:
                     field = dataclasses.replace(field, typ=type(field.default))
-                    assert is_possibly_nested_type(field.typ)
+                    assert _fields.is_nested_type(field.typ, field.default)
 
             # (2) Handle nested callables.
-            if is_possibly_nested_type(field.typ):
+            if _fields.is_nested_type(field.typ, field.default):
+                field = dataclasses.replace(
+                    field, typ=_resolver.narrow_type(field.typ, field.default)
+                )
                 nested_parser = ParserSpecification.from_callable(
                     field.typ,
                     description=None,
                     parent_classes=parent_classes,
                     parent_type_from_typevar=type_from_typevar,
                     default_instance=field.default,
-                    prefix=prefix + field.name + _strings.NESTED_FIELD_DELIMETER,
+                    prefix=_strings.make_field_name([prefix, field.name]),
                     avoid_subparsers=avoid_subparsers,
                 )
                 args.extend(nested_parser.args)
 
                 for k, v in nested_parser.helptext_from_nested_class_field_name.items():
                     helptext_from_nested_class_field_name[
-                        field.name + _strings.NESTED_FIELD_DELIMETER + k
+                        _strings.make_field_name([field.name, k])
                     ] = v
 
                 if field.helptext is not None:
@@ -214,9 +139,7 @@ class ParserSpecification:
                 else:
                     helptext_from_nested_class_field_name[
                         field.name
-                    ] = _docstrings.get_callable_description(
-                        _resolver.resolve_generic_types(field.typ)[0]
-                    )
+                    ] = _docstrings.get_callable_description(field.typ)
                 continue
 
             # (3) Handle primitive types. These produce a single argument!
@@ -275,11 +198,10 @@ class ParserSpecification:
                 continue
 
             if arg.prefix not in group_from_prefix:
-                nested_field_name = arg.prefix[:-1]
                 group_from_prefix[arg.prefix] = parser.add_argument_group(
-                    format_group_name(nested_field_name),
+                    format_group_name(arg.prefix),
                     description=self.helptext_from_nested_class_field_name.get(
-                        nested_field_name
+                        arg.prefix
                     ),
                 )
             arg.add_argument(group_from_prefix[arg.prefix])
@@ -288,51 +210,9 @@ class ParserSpecification:
         if len(self.subparsers_from_name) > 0:
             prev_subparser_tree_nodes = [parser]  # Root node.
             for subparsers in self.subparsers_from_name.values():
-                title = "subcommands"
-                metavar = "{" + ",".join(subparsers.parser_from_name.keys()) + "}"
-                if not subparsers.required:
-                    title = "optional " + title
-                    metavar = f"[{metavar}]"
-
-                subparser_tree_nodes = []
-                for p in prev_subparser_tree_nodes:
-                    # Add subparsers to every node in previous level of the tree.
-                    argparse_subparsers = p.add_subparsers(
-                        dest=self.prefix
-                        + _strings.SUBPARSER_DEST_FMT.format(name=subparsers.name),
-                        description=subparsers.description,
-                        required=subparsers.required,
-                        title=title,
-                        metavar=metavar,
-                    )
-                    for name, subparser_def in subparsers.parser_from_name.items():
-                        subparser = argparse_subparsers.add_parser(
-                            name,
-                            formatter_class=_argparse_formatter.ArgparseHelpFormatter,
-                        )
-                        subparser_def.apply(subparser)
-
-                        def _get_leaf_subparsers(
-                            node: argparse.ArgumentParser,
-                        ) -> List[argparse.ArgumentParser]:
-                            if node._subparsers is None:
-                                return [node]
-                            else:
-                                # Magic!
-                                return list(
-                                    itertools.chain(
-                                        *map(
-                                            _get_leaf_subparsers,
-                                            node._subparsers._actions[
-                                                -1
-                                            ]._name_parser_map.values(),  # type: ignore
-                                        )
-                                    )
-                                )
-
-                        subparser_tree_nodes.extend(_get_leaf_subparsers(subparser))
-
-                prev_subparser_tree_nodes = subparser_tree_nodes
+                prev_subparser_tree_nodes = subparsers.apply(
+                    self, prev_subparser_tree_nodes
+                )
 
 
 @dataclasses.dataclass(frozen=True)
@@ -342,6 +222,7 @@ class SubparsersSpecification:
     name: str
     description: Optional[str]
     parser_from_name: Dict[str, ParserSpecification]
+    prefix: str
     required: bool
     default_instance: Any
     can_be_none: bool  # If underlying type is Optional[Something].
@@ -361,30 +242,33 @@ class SubparsersSpecification:
         # We don't use sets here to retain order of subcommands.
         options = [type_from_typevar.get(typ, typ) for typ in get_args(field.typ)]
         options_no_none = [o for o in options if o != type(None)]  # noqa
-        if not all(map(is_possibly_nested_type, options_no_none)):
+        if not all(
+            [
+                _fields.is_nested_type(o, _fields.MISSING_NONPROP)
+                for o in options_no_none
+            ]
+        ):
             return None
 
+        # Add subparser for each option.
         parser_from_name: Dict[str, ParserSpecification] = {}
         for option in options_no_none:
-            subparser_name = _strings.subparser_name_from_type(option)
+            subparser_name = _strings.subparser_name_from_type(prefix, option)
             parser_from_name[subparser_name] = ParserSpecification.from_callable(
                 option,
                 description=None,
                 parent_classes=parent_classes,
                 parent_type_from_typevar=type_from_typevar,
                 default_instance=field.default
-                if isinstance(field.default, _resolver.unwrap_origin(option))
-                else None,
+                if type(field.default) == _resolver.unwrap_origin(option)  # type: ignore
+                else _fields.MISSING_NONPROP,
                 prefix=prefix,
                 avoid_subparsers=avoid_subparsers,
             )
 
         # Optional if: type hint is Optional[], or a default instance is provided.
         required = True
-        can_be_none = options != options_no_none  # Optional[] type.
-        if can_be_none:
-            required = False
-        elif field.default not in _fields.MISSING_SINGLETONS:
+        if field.default not in _fields.MISSING_SINGLETONS:
             required = False
 
         # If there are any required arguments in the default subparser, we should mark
@@ -393,8 +277,15 @@ class SubparsersSpecification:
             field.default is not None
             and field.default not in _fields.MISSING_SINGLETONS
         ):
+            # It's really hard to concretize a generic type at runtime, so we just...
+            # don't. :-)
+            if hasattr(type(field.default), "__parameters__"):
+                raise _instantiators.UnsupportedTypeAnnotationError(
+                    "Default values for generic subparsers are not supported."
+                )
+
             default_parser = parser_from_name[
-                _strings.subparser_name_from_type(type(field.default))
+                _strings.subparser_name_from_type(prefix, type(field.default))
             ]
             if any(map(lambda arg: arg.lowered.required, default_parser.args)):
                 required = True
@@ -406,23 +297,108 @@ class SubparsersSpecification:
             ):
                 required = True
 
+        # Make description.
         description_parts = []
         if field.helptext is not None:
             description_parts.append(field.helptext)
         if not required and field.default not in _fields.MISSING_SINGLETONS:
-            default = _strings.subparser_name_from_type(type(field.default))
+            default = field.default
+            if default is not None:
+                default = _strings.subparser_name_from_type(prefix, type(default))
             description_parts.append(f" (default: {default})")
+        description = (
+            # We use `None` instead of an empty string to prevent a line break from
+            # being created where the description would be.
+            " ".join(description_parts)
+            if len(description_parts) > 0
+            else None
+        )
 
         return SubparsersSpecification(
             name=field.name,
             # If we wanted, we could add information about the default instance
             # automatically, as is done for normal fields. But for now we just rely on
             # the user to include it in the docstring.
-            description=" ".join(description_parts),
+            description=description,
             parser_from_name=parser_from_name,
+            prefix=prefix,
             required=required,
-            default_instance=field.default
-            if field.default not in _fields.MISSING_SINGLETONS
-            else None,
-            can_be_none=can_be_none,
+            default_instance=field.default,
+            # if field.default not in _fields.MISSING_SINGLETONS
+            # else None,
+            can_be_none=options != options_no_none,
         )
+
+    def apply(
+        self,
+        parent_parser: ParserSpecification,
+        prev_subparser_tree_nodes: List[argparse.ArgumentParser],
+    ) -> List[argparse.ArgumentParser]:
+        title = "subcommands"
+        metavar = (
+            "{"
+            + ",".join(
+                (
+                    (_strings.subparser_name_from_type(self.prefix, None),)
+                    if self.can_be_none
+                    else ()
+                )
+                + tuple(self.parser_from_name.keys())
+            )
+            + "}"
+        )
+        if not self.required:
+            title = "optional " + title
+            metavar = f"[{metavar}]"
+
+        subparser_tree_nodes: List[argparse.ArgumentParser] = []
+        for p in prev_subparser_tree_nodes:
+            # Add subparsers to every node in previous level of the tree.
+            argparse_subparsers = p.add_subparsers(
+                dest=_strings.make_field_name(
+                    [
+                        parent_parser.prefix,
+                        _strings.make_subparser_dest(name=self.name),
+                    ]
+                ),
+                description=self.description,
+                required=self.required,
+                title=termcolor.colored(title, attrs=["bold"]),
+                metavar=metavar,
+            )
+
+            if self.can_be_none:
+                subparser = argparse_subparsers.add_parser(
+                    name=_strings.subparser_name_from_type(self.prefix, None),
+                    formatter_class=_argparse_formatter.ArgparseHelpFormatter,
+                )
+                subparser_tree_nodes.append(subparser)
+
+            for name, subparser_def in self.parser_from_name.items():
+                subparser = argparse_subparsers.add_parser(
+                    name,
+                    formatter_class=_argparse_formatter.ArgparseHelpFormatter,
+                )
+                subparser_def.apply(subparser)
+
+                def _get_leaf_subparsers(
+                    node: argparse.ArgumentParser,
+                ) -> List[argparse.ArgumentParser]:
+                    if node._subparsers is None:
+                        return [node]
+                    else:
+                        # Magic!
+                        return list(
+                            itertools.chain(
+                                *map(
+                                    _get_leaf_subparsers,
+                                    node._subparsers._actions[
+                                        -1
+                                    ]._name_parser_map.values(),  # type: ignore
+                                )
+                            )
+                        )
+
+                subparser_tree_nodes.extend(_get_leaf_subparsers(subparser))
+
+        return subparser_tree_nodes
