@@ -3,12 +3,35 @@
 import functools
 import re
 import textwrap
-from typing import Type
+from typing import Iterable, List, Sequence, Type, Union
 
 from . import _resolver
 
-NESTED_FIELD_DELIMETER: str = "."
-SUBPARSER_DEST_FMT: str = "{name} (positional)"
+dummy_field_name = "__dcargs_dummy_field_name__"
+
+
+def _strip_dummy_field_names(parts: Iterable[str]) -> Iterable[str]:
+    return filter(lambda name: len(name) > 0 and name != dummy_field_name, parts)
+
+
+def make_field_name(parts: Sequence[str]) -> str:
+    """Join parts of a field name together. Used for nesting.
+
+    ('parent', 'child') => 'parent.child'
+    ('parents', '1', 'child') => 'parents:1.child'
+    """
+    out: List[str] = []
+    for i, p in enumerate(_strip_dummy_field_names(parts)):
+        if i > 0:
+            out.append(".")
+
+        out.append(p)
+
+    return "".join(out)
+
+
+def make_subparser_dest(name: str) -> str:
+    return f"{name} (positional)"
 
 
 def dedent(text: str) -> str:
@@ -28,7 +51,7 @@ def hyphen_separated_from_camel_case(name: str) -> str:
     return _camel_separator_pattern().sub(r"-\1", name).lower()
 
 
-def subparser_name_from_type(cls: Type) -> str:
+def _subparser_name_from_type(cls: Type) -> str:
     cls, type_from_typevar = _resolver.resolve_generic_types(cls)
     if len(type_from_typevar) == 0:
         assert hasattr(cls, "__name__")
@@ -36,12 +59,19 @@ def subparser_name_from_type(cls: Type) -> str:
 
     return "-".join(
         map(
-            subparser_name_from_type,
+            _subparser_name_from_type,
             [cls] + list(type_from_typevar.values()),
         )
     )
 
 
+def subparser_name_from_type(prefix: str, cls: Union[Type, None]) -> str:
+    suffix = _subparser_name_from_type(cls) if cls is not None else "None"
+    if len(prefix) == 0:
+        return suffix
+    return f"{prefix}:{suffix}"
+
+
 @functools.lru_cache(maxsize=None)
 def _get_ansi_pattern() -> re.Pattern:
     # https://stackoverflow.com/a/14693789
